var map, markers, loadingPanel;
var MAX_ITEMS = 100;
var DELIMITER = "\n\n"; // TODO change ?
var MAP_TYPES = null;
var features = new Object();
var GEONAMES = new Object();;
var IGNORED_WOEIDS = new Object();
var QUERY_SPLIT_LEN = 45000;
var QUERY_WORD_OVERLAP = 10;
var ERROR_MESSAGE = '';

var localeBundle = initLocaleBundle();

var MODULE_NAME = "ignored";
var FIRST_STATE = YAHOO.util.History.getBookmarkedState(MODULE_NAME) || "";
YAHOO.util.History.register(MODULE_NAME, FIRST_STATE, on_url_change);


// NOTE: "tags", "notes", and "attachments" are not included
// because they require special handlers 
var ITEM_FIELDS = [
    "place",
    "archiveLocation",
    "court",
    "institution",
    "university",

    "title",
    "abstractNote",
    "extra"
];

var items_to_load = 0;
var llproj = new OpenLayers.Projection("EPSG:4326");
var googproj = new OpenLayers.Projection("EPSG:900913");


function initLocaleBundle() {
    var src = 'chrome://zotero-maps/locale/maps.properties';
    var localeService = Components.classes['@mozilla.org/intl/nslocaleservice;1']
                                  .getService(Components.interfaces.nsILocaleService);
    var appLocale = localeService.getApplicationLocale();
    
    var stringBundleService = Components.classes["@mozilla.org/intl/stringbundle;1"]
                                        .getService(Components.interfaces.nsIStringBundleService);
    return stringBundleService.createBundle(src, appLocale);
}

/*
 * Get a localized string from the string bundle
 *
 * This is copied from zotero.js
 */
function getString(name, params){
try {
        if (params != undefined){
            if (typeof params != 'object'){
                params = [params];
            }
            var l10n = localeBundle.formatStringFromName(name, params, params.length);
        }
        else {
            var l10n = localeBundle.GetStringFromName(name);
        }
    }
    catch (e){
        throw ('Localized string not available for ' + name);
    }
    return l10n;
}




var Feature = OpenLayers.Class(OpenLayers.Feature, {
    popupClass: OpenLayers.Popup.FramedCloud,
    mouseDown: function (evt) {
        var existing = map.popups[0];
        if (existing && existing.feature == this) {
            existing.toggle();
        } else { 
           
            var popup = this.createPopup(true);
            var html = '<b>'+this.data.name+'</b> ('+this.data.items.length+')<hr />';
            for (var i = 0; i<this.data.items.length; i++) {
                var item = this.data.items[i];
                /* FIXME: this is really ugly and should be refactored somehow */
                html +='<div style="margin-left:24px; font-size:small">'
                        +'<img src="' + item.typeIcon + '" style="margin-left:-24px; margin-top:4px;" /> '
                        +'<a href="zotero://select/item/' + item.id + '">'
                        + item.title + '</a>'
                        + '</div>';
            }

            // add the 'Remove Marker from Map' option to the popup
            html += '<div style="margin-left:24px; font-size:small">'
                     +'<img src="chrome://zotero-maps/skin/delete.png" style="margin-left:-24px; margin-top:4px;" /> '
                     +'<a href="" onclick="remove_marker(' + this.data.woeId + '); return false;">'
                     + getString("marker.remove") + '</a></div>';

            popup.feature = this;
            popup.setContentHTML(html);
            map.addPopup(popup,true);
            popup.show();
        }
        OpenLayers.Event.stop(evt);
        return true;
    } 
});

function get_items_from_zotero () {
    var path = window.location.pathname.split("/");
    var type = path[2];
    var ids  = path[3];
    var results;
    switch (type) {
        case 'collection':
            var col = Zotero.Collections.get(ids);
            results = col.getChildItems();
            break;
        case 'search':
            try {
                var s = new Zotero.Search(ids); // Zotero 1.*
            }
            catch(e) {
                // Zotero 2.*
                var s = new Zotero.Search(); 
                s.id = ids;
            }
            ids = s.search();
            break;
        case 'group':
            var s = new Zotero.Search();
            s.addCondition('noChildren', 'true');
            s.addCondition('libraryID', 'is', ids);
            ids = s.search();
            break;
        default:
            type = 'library';
            var s = new Zotero.Search();
            s.addCondition('noChildren', 'true');
            s.addCondition('libraryID', 'is', null);
            ids = s.search();
    }
    if (!results) {
        var results = Zotero.Items.get(ids);
    }
    var items = [];
    // Only include parent items
    for (var i = 0; i < results.length; i++) {
        if (!results[i].getSource()) {
            items.push(results[i]);
        }
    }
     
    return items;
}  

function add_item_to_features (item, geoname) {
    var loc = new OpenLayers.LonLat(geoname.lng, geoname.lat);
    var key = geoname.woeId; //loc.toShortString();
    loc.transform(llproj,googproj);

    /* Key Zotero items by lat/lon so that citations can be
     * grouped by location. */

    if (!features[key]) {
        var size = new OpenLayers.Size(16,28); 
        var offset = new OpenLayers.Pixel(-(size.w/2), -size.h); 
        var zoteroIcon = new OpenLayers.Icon('chrome://zotero-maps/skin/img/zotero_16x28.png',size,offset);
        var feature = new Feature(markers, loc, {name:geoname.name, woeId: geoname.woeId, items:[], icon:zoteroIcon.clone()});
        var marker = feature.createMarker();
        markers.addMarker(marker);
        marker.events.register("mousedown", feature, feature.mouseDown);
        features[key] = feature;
    }
    var citation = {
       title: item.getDisplayTitle(),
       id: item.getID(),
       type: item.getType(),
       typeIcon: item.getImageSrc()
    };
    features[key].data.items.push(citation);
};

// return a string containing the text that Yahoo
// will search for the given item
function process_item(item) {
    var itemStr = "";

    for(var i = 0, len = ITEM_FIELDS.length; i < len; i++) {
        if(MAP_TYPES[ITEM_FIELDS[i]]) {
            var fieldVal = item.getField(ITEM_FIELDS[i]);
            if(fieldVal) {
                itemStr += fieldVal + DELIMITER;
            }
        }
    }

    if(MAP_TYPES.tags) {
        var tags = item.getTags();
        for(var i = 0, len = tags.length; i < len; i++) {
            itemStr += tags[i].name + DELIMITER;
        }
    }

    if(MAP_TYPES.notes) {
        if(item.isNote()) {
            var noteText = item.getNote();
            if(noteText) {
                itemStr += Zotero.Utilities.unescapeHTML(noteText) + DELIMITER;
            }
        }
        else {
            var notes = Zotero.Items.get(item.getNotes());
            for(var i = 0, len = notes.length; i < len; i++) {
                itemStr += process_item(notes[i]);
            }
        }
    }

    if(MAP_TYPES.attachments) {
        if(item.isAttachment()) {
            var attachmentText = item.attachmentText;
            if(attachmentText) {
                itemStr += item.attachmentText + DELIMITER;
            }
        }
        else {
            var attachments = Zotero.Items.get(item.getAttachments());
            for(var i = 0, len = attachments.length; i < len; i++) {
                itemStr += process_item(attachments[i]);
            }
        }
    }

    return itemStr;
}


function populate_map () {
    var documentContent = "";

    // keep track of the start and end of the searchable text for each item
    // in order to correspond locations with items (yahoo returns the found
    // locations, along with the start and end indexes of the location text)
    var itemBounds = new Array();

    // items that have searchable text are pushed into this array
    // so that itemBounds[i] refers to usedItems[i]
    var usedItems = new Array();

    var items = get_items_from_zotero();
    var itemsLen = items.length;
    if (itemsLen > MAX_ITEMS) {
        itemsLen = MAX_ITEMS;
        alert(getString("alert.moreThanMax", [itemsLen, MAX_ITEMS]));
    }

    for(var j=0; j<itemsLen; j++) {
        var itemStr = process_item(items[j]);
        if(itemStr == "") {
            continue;
        }
        
        usedItems.push(items[j]);
        itemBounds.push({
            start: documentContent.length,
            end: documentContent.length + itemStr.length - 1
        });

        documentContent += itemStr; // ending DELIMITER added by process_item
    }
    
    if(documentContent == "") {
        set_map_zoom();
        return;
    }

    var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                          .getService(Components.interfaces.nsIPrefBranch);

    // Yahoo appID registered to zotero.org for maps plugin
    var appid = "lg5BJ67V34HbQ0iL3nSRNAfYMe264QTlf4Yp6vjvnvF0PYqGcWtWKReHoqMy";
    var documentType = "text/plain";
    
    // set inputLanguage to user's locale
    var inputLanguage = prefs.getCharPref("general.useragent.locale");

    // split the request into parts so we don't hit Yahoo's character limit
    var documentContents = []; // contains the parts

    // keep track of the string offsets between the new smaller requests
    // and the original full request which the itemBounds were based off of
    var boundOffsets = [];
    var currentBoundOffset = 0;

    var splitWordRegExp = new RegExp("^\\S+");
    var overlapRegExp = new RegExp("^\\s*(\\S+\\s+){" + QUERY_WORD_OVERLAP + "}");
    while(documentContent.length > QUERY_SPLIT_LEN) {
        boundOffsets.push(currentBoundOffset);

        var part = documentContent.substring(0, QUERY_SPLIT_LEN);
        documentContent = documentContent.substring(QUERY_SPLIT_LEN);
        
        // could have split in the middle of a word, so add
        // the last part of the word
        var splitWord = documentContent.match(splitWordRegExp);
        if(splitWord) {
            part += splitWord[0];
            documentContent = documentContent.substring(splitWord[0].length);
        }
        
        currentBoundOffset += part.length;

        // add overlapping words to ensure we don't split in the middle of a place
        var overlap = documentContent.match(overlapRegExp);
        if(overlap) {
            part += overlap[0];
        }
        
        documentContents.push(part);
    }
    if(documentContent.length > 0) {
        boundOffsets.push(currentBoundOffset);
        documentContents.push(documentContent);
    }

    // complete requests, and store the req.responseXML of each in the domTrees array
    var requestCount = documentContents.length;
    var requestsComplete = 0;
    var domTrees = new Array(requestCount);
    for(var i = 0; i < requestCount; i++) {
        var post = "appid=" + appid +
                   "&documentContent=" + encodeURIComponent(documentContents[i]) +
                   "&documentType=" + documentType +
                   "&inputLanguage=" + inputLanguage;

        var req = new XMLHttpRequest();
        req.open('POST', 'http://wherein.yahooapis.com/v1/document', true);
        req.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
        req.requestNumber = i;
        req.onreadystatechange = function (aEvt) {
            var req = aEvt.originalTarget;
            if (req.readyState == 4) {
                requestsComplete++;

                if(req.status == 200) {
                    domTrees[req.requestNumber] = req.responseXML;
                    if(requestsComplete == requestCount) {
                        // if all requests have completed, process the dom trees
                        process_dom_trees(domTrees, usedItems, itemBounds, boundOffsets);
                    }
                }
                else {
                    ERROR_MESSAGE += getString("alert.error", req.status) + '\n';
                }
            }
        };
        req.send(post);
    }
}

function process_dom_trees(domTrees, items, itemBounds, boundOffsets) {
    // only show one error, instead of one per errored request
    if(ERROR_MESSAGE) {
        alert(ERROR_MESSAGE);
    }

    var itemsLen = items.length;
    for(var i = 0, len = domTrees.length; i < len; i++) {
        if(!domTrees[i]) {
            // skip if server errored on this request
            continue;
        }

        var places = domTrees[i].getElementsByTagName("placeDetails");
        var references = domTrees[i].getElementsByTagName("reference");
        var boundOffset = boundOffsets[i];

        for(var j = 0, len2 = places.length; j < len2; j++) {
            var woeId = places[j].getElementsByTagName("woeId")[0].textContent;
            if(!GEONAMES[woeId]) {
                GEONAMES[woeId] = {
                    woeId: woeId,
                    lng: places[j].getElementsByTagName("longitude")[0].textContent,
                    lat: places[j].getElementsByTagName("latitude")[0].textContent,
                    name: places[j].getElementsByTagName("name")[0].textContent,
                    items: new Object()
                };
            }
        }
        for(var j = 0, len2 = references.length; j < len2; j++) {
            var woeId = references[j].getElementsByTagName("woeIds")[0].textContent.split(" ", 1)[0];

            var start = references[j].getElementsByTagName("start")[0].textContent;
            start = parseInt(start) + boundOffset;
            var end = references[j].getElementsByTagName("end")[0].textContent;
            end = parseInt(end) + boundOffset;

            // binary search the bounds
            var minIndex = 0, maxIndex = itemsLen - 1, midIndex;
            while(minIndex <= maxIndex) {
                midIndex = Math.floor((minIndex + maxIndex) / 2);
                if(start < itemBounds[midIndex].start) {
                    maxIndex = midIndex - 1;
                }
                else if(end  > itemBounds[midIndex].end) {
                    minIndex = midIndex + 1;
                }
                else {
                    // item found (location bound is enclosed in item bound)
                    break;
                }
            }

            // ensure the corresponding item was found
            // It's possible that Yahoo "found" a location whose text
            // spanned more than one item bound, and therefore is not
            // an actual found location
            if(minIndex > maxIndex) {
                continue;
            }
            
            var item = items[midIndex];

            // ensure an item is only added to a location once
            if(!GEONAMES[woeId].items[item.id]) {
                GEONAMES[woeId].items[item.id] = true;
                add_item_to_features(item, GEONAMES[woeId]);
            }
        }
    }

    // update displayed markers (hide markers in original list from url)
    on_url_change(FIRST_STATE);

    set_map_zoom();
}

//
function remove_marker(woeId) {
    var newState = YAHOO.util.History.getCurrentState(MODULE_NAME);
    if(newState) {
        // add delimeter if woeId is not the first in the list
        newState += "-";
    }
    newState += woeId;

    // navigate to new state triggering call to on_url_change
    // Note: by using this Yahoo library, this is done without a page reload
    YAHOO.util.History.navigate(MODULE_NAME, newState);
}


// find changes to the state and change corresponding display states
// state contains a '-' delimited list of woeIds that we ignore (don't display)
function on_url_change(state) {
    // hide popup if exists (ensure a popup to a hidden marker won't happen)
    var popup = map.popups[0];
    if (popup) {
        popup.hide();
    }

    var new_IGNORED_WOEIDS = new Object();
    var woeIds = state.split("-");
    for(var i = 0, len = woeIds.length; i < len; i++) {
        var woeId = woeIds[i];
        if(woeId) {
            if(!IGNORED_WOEIDS[woeId]) {
                // woeId was added to the list, so should be not be displayed
                features[woeId].marker.display(false);
            }
            new_IGNORED_WOEIDS[woeId] = true;
        }
    }

    for(woeId in IGNORED_WOEIDS) {
        if(!new_IGNORED_WOEIDS[woeId]) {
            // woeId was removed from the list, so should be displayed
            features[woeId].marker.display(true);
        }
    }

    IGNORED_WOEIDS = new_IGNORED_WOEIDS;
}


function set_map_zoom() {
    var bounds = markers.getDataExtent();
    if (bounds) {
       map.zoomToExtent(bounds);
       if (markers.markers.length == 1) {
            map.zoomTo(4);
       }
    }else{
       map.zoomToMaxExtent();
       map.zoomTo(2);
    }
}

function osm_getTileURL(bounds) {
    var res = this.map.getResolution();
    var x = Math.round((bounds.left - this.maxExtent.left) / (res * this.tileSize.w));
    var y = Math.round((this.maxExtent.top - bounds.top) / (res * this.tileSize.h));
    var z = this.map.getZoom();
    var limit = Math.pow(2, z);
 
    if (y < 0 || y >= limit) {
        return OpenLayers.Util.getImagesLocation() + "blank.gif";
    } else {
        x = ((x % limit) + limit) % limit;
        return this.url + z + "/" + x + "/" + y + "." + this.type;
    }
}

function onLoad() {
    // avoid pink tiles
    OpenLayers.IMAGE_RELOAD_ATTEMPTS = 3;
    OpenLayers.Util.onImageLoadErrorColor = "white";

    loadingPanel = new OpenLayers.Control.LoadingPanel();
    var options = {
          projection: googproj,
          displayProjection: llproj,
          units: "m",
          maxResolution: 156543,
          maxExtent: new OpenLayers.Bounds(-20037508, -20037508, 20037508, 20037508),
          numZoomLevels: 12,
          controls: [
            new OpenLayers.Control.PanZoomBar(),
            new OpenLayers.Control.KeyboardDefaults(),
            new OpenLayers.Control.Navigation(),
            new OpenLayers.Control.Attribution(),
            loadingPanel
          ]
          //style:null
    };

    map = new OpenLayers.Map('map', options);

    /*** Google Maps won't work without a license key...
    * and Google doesn't give them out for zotero:// or chrome:// APIs...
    var gmap = new OpenLayers.Layer.Google( "Map" , {
        type: G_NORMAL_MAP, sphericalMercator:true,'maxZoomLevel':12});
    */

    // create OSM layer
    var mapnik = new OpenLayers.Layer.TMS(
        "OpenStreetMap (Mapnik)",
        "http://tile.openstreetmap.org/",
        {
            type: 'png', getURL: osm_getTileURL,
            displayOutsideMaxExtent: true,
            attribution: 'Map provided by ' +
                    '<a href="http://www.openstreetmap.org/">OpenStreetMap</a>',
            transitionEffect: 'resize'
            //wrapDateLine: true
        }
    );

    markers = new OpenLayers.Layer.Markers("markers");
    map.addLayers([mapnik, markers]);
    
      loadingPanel.setVisible(false);

    //turn off default layer event stuff for the loadingPanel...hack.
    for (var i = 0; i < map.layers.length; i++) {
            var layer = map.layers[i];
            layer.events.unregister('loadstart', loadingPanel);
            layer.events.unregister('loadend', loadingPanel);
        }


    // process map types
    var mapTypeArray = window.location.pathname.split("/")[1].split("+");
    MAP_TYPES = new Object();
    for(var i = 0, len = mapTypeArray.length; i < len; i++) {
        MAP_TYPES[mapTypeArray[i]] = true;
    }

    // initialize Browser History Manager
    // "yui-history-field" = id of hidden input field in ui.html
    YAHOO.util.History.initialize("yui-history-field");

    populate_map();
}
